#version 410 core

uniform sampler2D position_sampler;
uniform sampler2D diffuse_sampler;
uniform sampler2D normal_sampler;
uniform sampler2D specular_sampler;
uniform samplerCube shadow_sampler;

uniform vec3 camera_pos;
uniform vec3 light_pos_deferred;
uniform vec2 screen_size;

uniform vec3 light_color;

uniform bool shadowmapping;

uniform float far_plane;

uniform float ambient_intensity;
uniform float diffuse_intensity;
uniform float specular_intensity;
uniform float specular_power;
uniform float decay;

uniform float atten_con;
uniform float atten_lin;
uniform float atten_exp;

layout (location = 4) out vec4 bright_color;
layout (location = 5) out vec4 frag_color;

const vec3 sample_offset_directions[20] = vec3[] (
   vec3( 1,  1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1,  1,  1), 
   vec3( 1,  1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1,  1, -1),
   vec3( 1,  1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1,  1,  0),
   vec3( 1,  0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1,  0, -1),
   vec3( 0,  1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0,  1, -1)
);

float calculate_shadow_factor(vec3 frag_to_light_vec) {
	float shadow = 0.0;
	float eps = 0.15;
	int samples = 20;
	float disk_radius = 0.01;
	
	float current_depth = length(frag_to_light_vec);
	frag_to_light_vec = normalize(frag_to_light_vec);
	
	for(int i = 0; i < samples; ++i) {
		float closest_depth = texture(shadow_sampler, -frag_to_light_vec + sample_offset_directions[i] * disk_radius).r;
		closest_depth *= far_plane;
		
		if(current_depth - eps > closest_depth)
			shadow += 1.0;
	}
	return shadow / float(samples);
}

void main(void) {
	// get the uv coordinate
	vec2 uv = gl_FragCoord.xy / screen_size;

	// do the lookups in the textures
	vec3 position_texture = texture(position_sampler, uv).xyz;
	vec3 diffuse_texture = texture(diffuse_sampler, uv).xyz;
	vec3 normal_texture = texture(normal_sampler, uv).xyz;
	vec3 specular_texture = texture(specular_sampler, uv).xyz;
	
	// do the lightning calculations
	vec3 light_dir = light_pos_deferred - position_texture;
	float distance = length(light_dir);
	float shadow_factor = shadowmapping ? calculate_shadow_factor(light_dir) : 0.0f;
	light_dir = normalize(light_dir);
	
	// ambient term
	vec3 ambient_color = light_color * ambient_intensity;
	
	// diffuse term
	float diffuse_factor = max(dot(light_dir, normalize(normal_texture)), 0);
	vec3 diffuse_color = light_color * diffuse_intensity * diffuse_factor;
	
	// specular term
	vec3 view_dir = normalize(camera_pos - position_texture);
	vec3 half_vec = normalize(light_dir + view_dir);
	float specular_factor = pow(max(dot(half_vec, normalize(normal_texture)), 0), specular_power);
	//vec3 specular_color = specular_texture * specular_intensity * specular_factor;
	vec3 specular_color = specular_texture * specular_factor;
	
	vec3 current_color = diffuse_texture * (
		ambient_color +
		max(0.0, (1.0f - shadow_factor)) * (diffuse_color + specular_color)
		);

	// calculate the attenuation
	float attenuation = max(1.0, atten_con + atten_lin * distance + atten_exp * distance * distance);
	current_color = current_color / attenuation;

	float brightness = dot(current_color, vec3(0.2126f, 0.7152f, 0.0722f));
    if(brightness > 1.0f) {
        bright_color = vec4(current_color, 1.0);
	} else {
		bright_color = vec4(0,0,0,1);
	}
	frag_color = vec4(current_color, 1) * (1.0 - decay);
	
	//frag_color = vec4(vec3(abs(distance-(texture(shadow_sampler, -light_dir).r) * far_plane)), 1.0); 
	//frag_color = vec4(vec3(texture(shadow_sampler, -light_dir).r), 1);
	//frag_color = vec4(vec3(distance / 25.0f - (texture(shadow_sampler, -light_dir).r)), 1);
	//frag_color = vec4(vec3(distance / 25.0f), 1);
	//frag_color = vec4(vec3(specular_texture.r / 25.0f), 1);

}