#version 410 core

in vec2 uv;
in vec3 frag_pos;
in vec3 frag_pos_ts;
in vec3 camera_pos_ts;
in mat3 tbn_mat;
in mat3 tbn_mat_inv;

layout (location = 0) out vec3 position_out;
layout (location = 1) out vec3 diffuse_out;
layout (location = 2) out vec3 normal_out;
layout (location = 3) out vec3 specular_out;

uniform sampler2D diffuse_sampler;
uniform sampler2D normal_sampler;
uniform sampler2D specular_sampler;
uniform sampler2D displacement_sampler;
uniform vec3 light_pos;
uniform float bounding_scale;

uniform vec3 camera_pos;

uniform bool parallax;
uniform bool is_light_source;

void main(void) {
	vec2 final_uv = uv;
	
	// calculate distance to light source
	float distance = length(frag_pos - light_pos);
	
	if(!is_light_source && parallax) {
		// perform parallax mapping
		vec3 view_dir_ts = normalize(camera_pos_ts - frag_pos_ts);
		
		const float min_layers = 1;
		const float max_layers = 32;
		const float height_scale = 0.05f;

		// num_layers is a mixture from the distance to the light source and the viewing direction
		// near a light source the samples are calculated from the viewing angle
		// beginning at 1/4 distance to the bounding sphere the samples are decreasing
		float num_layers;
		if(distance >= bounding_scale / 4.0f)
			num_layers = mix(1, 32, 4.f * distance / bounding_scale);
		else
			num_layers = mix(32, 16, abs(dot(vec3(0.0, 0.0, 1.0), view_dir_ts)));
		
		float layer_depth = 1.0 / num_layers;
		float current_layer_depth = 0.0;
		
		vec2 delta_uv = (view_dir_ts.xy / view_dir_ts.z * height_scale) / num_layers;
		vec2 current_uv = uv;
		float current_displacement = texture(displacement_sampler, current_uv).r;
		
		float counter = 0;
		while(current_layer_depth < current_displacement && counter <= num_layers) {
			current_uv -= delta_uv;
			current_displacement = texture(displacement_sampler, current_uv).r; 
			current_layer_depth += layer_depth;  
			counter++;
		}
		
		// parallax occlusion mapping
		vec2 prev_uv = current_uv + delta_uv;
		float after_depth  = current_displacement - current_layer_depth;
		float before_depth = texture(displacement_sampler, prev_uv).r - current_layer_depth + layer_depth;
		float weight = after_depth / (after_depth - before_depth);
		final_uv = prev_uv * weight + current_uv * (1.0 - weight);
	}
	

	// write to textures...
	position_out = frag_pos;
		
	if(is_light_source) {
		// if the color should not be hardcoded the particle must have a texture
		diffuse_out = texture(diffuse_sampler, final_uv).rgb * 3.0f; // bloom it
		normal_out = normalize(frag_pos - camera_pos);
		specular_out = vec3(0,0,0);
	} else {
		diffuse_out = texture(diffuse_sampler, final_uv).rgb;
		normal_out = texture(normal_sampler, final_uv).rgb;
		normal_out = normalize(normal_out * 2.0 - 1.0);
		normal_out = normalize(tbn_mat * normal_out);
		specular_out = texture(specular_sampler, final_uv).rgb;
	}
}
